/*
 * Copyright (c) 2004-2012 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/* ----------------------------------------------------------------------
 * Functions to init and read the angle from  a KVH5000 gyro
 * ---------------------------------------------------------------------- */


#include <err.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>

#include "serial.h"

#include "gyroPrivate.h"
#include "KVH5000.h"

static void *gyroKVH5000ReadTask(void *);
static int gyroClose(struct GYRO_DATA *);

/* ----------------------------------------------------------------------
 *
 * gyroInit -     Init connection to serial port
 *
 *
 * Returns :            NULL if failed or an initialized structure
 */

struct GYRO_DATA *
gyroInitKVH5000(const char *port, double latitude, double woffset, int sign)
{
	struct GYRO_KVH5000_DATA *gyroData;
	int rv;

	gyroData = malloc(sizeof(struct GYRO_KVH5000_DATA));
	if (gyroData == NULL)
		return NULL;
	gyroData->gd.lock = malloc(sizeof(pthread_mutex_lock));
	if (gyroData->gd.lock == NULL) {
		free(gyroData);
		return NULL;
	}
	pthread_mutex_init(gyroData->gd.lock, NULL);
	pthread_mutex_lock(gyroData->gd.lock);
	if ((gyroData->fd = openSerial(port, 38400)) == -1) {
		warn("openSerial %s", port);
		return NULL;
	}
	gyroData->gd.wOffset = woffset;
	gyroData->gd.wsign = sign;
	gyroData->gd.earthSpeed = gyroEarthSpeed(latitude);
	gyroData->gd.close = gyroClose;
	gyroData->stop = 0;
	rv = pthread_create(&gyroData->taskId, NULL,
	    gyroKVH5000ReadTask, gyroData);
	pthread_mutex_unlock(gyroData->gd.lock);
	if (rv != 0) {
		pthread_mutex_destroy(gyroData->gd.lock);
		free(gyroData);
		return NULL;
	}
	return (struct GYRO_DATA *)gyroData;
}


/*
 * gyroClose ------------------------------------------------------------
 */
static int
gyroClose(struct GYRO_DATA *p)
{
	struct GYRO_KVH5000_DATA *gyro = (struct GYRO_KVH5000_DATA *)p;

	gyro->stop = 1;
	if (close(gyro->fd) == -1)
		return -1;
	pthread_mutex_unlock(gyro->gd.lock);
	free(gyro->gd.lock);
	free(gyro);
	return 0;
}


/*----------------------------------------------------------------------*/

static void
gyroDecode(struct GYRO_KVH5000_DATA *gyro, char *buf, uint32_t date)
{
	double w;
	int val;

	if (strlen(buf) < 20) {
		warnx("gyroDecode: short frame");
		return;
	}
	if (sscanf(buf, "%lf %d", &w, &val) != 2) {
		warnx("gyroDecode: invalid frame");
		return;
	}

	w =  DEG_TO_RAD(gyro->gd.wsign*w) - gyro->gd.earthSpeed;

	pthread_mutex_lock(gyro->gd.lock);
	if (!val) {
		gyro->dataValid = 0;
		goto done;
	}
	gyro->gd.date = date;
	gyro->dataValid = 1;
	gyroIntegrate((struct GYRO_DATA *)gyro, w);
done:
	pthread_mutex_unlock(gyro->gd.lock);
}


static void *
gyroKVH5000ReadTask(void *param)
{
	static char buf[32];
	struct GYRO_KVH5000_DATA *gyro = (struct GYRO_KVH5000_DATA *)param;
	ssize_t nread;
	int rptr, rem;
	char *p;

	rptr = 0;
	while (!gyro->stop) {
		/* read some characters */
		nread = read(gyro->fd, buf + rptr, sizeof(buf) - rptr);
		if (nread < 0)
			err(2, "read");
		/* look for CR */
		p = strchr(buf+rptr, '\r');
		if (p != NULL) {
			/* found eom */
			*p = '\0';
			gyroDecode(gyro, buf, getTime());
			/* Handle remaining characters */
			rem = nread - (p - (buf + rptr)) - 1;
			/* recopy them at begining of buf */
			memmove(buf, p+1, rem);
			/* and clear the rest */
			memset(buf+rem, '0', sizeof(buf) - rem);
			rptr = rem;
		} else {
			rptr +=  nread;
			if (rptr == sizeof(buf)) {
				warnx("can't find EOM");
				rptr = 0;
			}
		}
	}
	return NULL;
}
