/*
 * Copyright (c) 2004,2012 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

#include <err.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>

#include <gyroLib/gyro.h>
#include <gyroLib/gyroTypes.h>

#define DEG_TO_RAD(x) ((x) * M_PI / 180.)
#define RAD_TO_DEG(x) ((x) * 180. / M_PI)

/* ----------------------------------------------------------------------
 *
 * gyroTest - init gyro, read and print angle
 *
 */
struct GYRO_DATA *gyro;
int stop = 0;

void
usage(void)
{
	errx(2, "usage: gyroTest [-c iters][-l latitude (deg)]\n"
	    "\t\t\t[-o woffset][-n iters] -1|-3|-5 <device>");
}

void
stop_handler (int sig)
{
	stop = 1;
}

static void
taskDelay(int period)
{
	struct timespec req, rem;
	int rv;

	req.tv_sec = period / 1000;
	req.tv_nsec = (period % 1000) * 1000000;

	for (;;) {
		rv = nanosleep(&req, &rem);
		if (rv == -1 && errno == EINTR) {
			memcpy(&req, &rem, sizeof(struct timespec));
			continue;
		}
		break;
	}
}


int
main(int argc, char *argv[])
{
	double theta, w;
	uint32_t date;
	int ch, period = 1000;	/* ms */
	char *dev;
	enum GYRO_TYPE type = GYRO_UNKNOWN;
	int i, num_calib = 0, num_iter = 0;
	double latitude = DEG_TO_RAD(43.56); /* LAAS */
	double woffset = 0.0;

	while ((ch = getopt(argc, argv, "c:l:n:o:135")) != -1) {
		switch (ch) {
		case 'c':
			/* calibration mode */
			num_calib = atoi(optarg);
			break;
		case 'l':
			latitude = strtod(optarg, NULL);
			break;
		case 'n':
			/* number of iterations to show, 0 - infinite loop */
			num_iter = atoi(optarg);
			break;
		case 'o':
			woffset = strtod(optarg, NULL);
			break;
		case '1':
			type = GYRO_KVH_ECORE_1000;
			period = 100;
			break;
		case '3':
			type = GYRO_KVH_DSP_3000;
			period = 10;
			break;
		case '5':
			type = GYRO_KVH_DSP_5000;
			period = 10;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();
	dev = argv[0];

	//printf ("Init with %s\n", dev);

	/* initialize gyroLib */
	gyro = gyroInit(type, dev, latitude, woffset);

	if (gyro == NULL)
		errx(2, "gyroInit: could not initialize gyro");

	if (num_calib != 0) {
		gyroUpdateWOffset(gyro, num_calib);
		printf("Estimated w offet: %g rad/s\n", gyroReadWOffset(gyro));
	}

	signal(SIGINT, stop_handler);

	i = 0;
	while (!stop) {

		taskDelay(period);

		/* read gyro */
		if (gyroRead(gyro, &theta, &w, &date) != 0)
			warnx("gyroTest: gyroRead failed");

		printf("%8d %.5g %.5g %.5g %.5g \n", date,
		    RAD_TO_DEG(theta), theta, RAD_TO_DEG(w), w);
		i++;
		if (num_iter != 0 && i == num_iter)
			break;
	}
	gyroEnd(gyro);
	printf ("bye\n");
	return 0;
}
